__version__ = "0.3.0-alpha.1"

from .bundle import Bundle  # noqa F401
from .dtn_rest_client import DTNRESTClient  # noqa F401
from .utils import from_dtn_timestamp, to_dtn_timestamp, RUNNING_MICROPYTHON  # noqa F401

if not RUNNING_MICROPYTHON:
    from .dtn_ws_client import DTNWSClient, WSMode  # noqa F401



import json
from base64 import b64encode
from enum import Enum
from typing import Any, Callable, ClassVar, Iterable, List, Optional, Union
from urllib import request as rq

import cbor2 as cbor
from websocket import ABNF, WebSocket, WebSocketApp

from py_dtn7 import Bundle


def _has_valid_schema(host: str):
    return host.startswith("wss://") or host.startswith("ws://")


def _rq_get(url: str) -> Any:
    return rq.urlopen(url=url)


class WSMode(Enum):
    DATA_MODE = 0
    JSON_MODE = 1


class DTNWSClient:
    """WebSocket Client connecting to a running dtnd instance"""

    # base ws endpoint URL
    _WS_BASE: ClassVar[str] = "/ws"
    # returns the node id of the local instance
    _NODE_ID_ENDPOINT: ClassVar[str] = "/node"
    # receive incoming bundles for this endpoint via the current websocket.
    # NOTE: the endpoint must be already registered to subscribe to it!
    _SUBSCRIBE_ENDPOINT: ClassVar[str] = "/subscribe"
    # stop receiving bundles for the given endpoint on this websocket connection.
    # NOTE: They are still collected on the node itself unless the endpoint is also unregistered!
    _UNSUBSCRIBE_ENDPOINT: ClassVar[str] = "/unsubscribe"
    # put this websocket into cbor data mode.
    _DATA_MODE: ClassVar[str] = "/data"
    # put this websocket into json mode.
    _JSON_MODE: ClassVar[str] = "/json"
    # put this websocket into raw bundle mode.
    _BUNDLE_MODE: ClassVar[str] = "/bundle"

    # instance attributes:
    _port: str
    _running: bool
    _callback: Union[Callable[[Bundle], Any], Callable[[str], Any]]
    _ws_base_url: str
    _endpoints: List[str]
    _mode: WSMode
    _ws: WebSocketApp

    def __init__(
        self,
        callback: Union[Callable[[Bundle], Any], Callable[[str], Any]],
        host: Optional[str] = None,
        port: Optional[str] = None,
        ws_base_url: Optional[str] = None,
        endpoints: Optional[Iterable[str]] = None,
    ):
        """

        :param callback: method to call when data is received
        :param host: host of DTN7 daemon
        :param port: port of DTN7 daemon
        """
        if port is None:
            port = 3000
        if host is None:
            host = "ws://localhost"
        if ws_base_url is None:
            ws_base_url = self._WS_BASE
        if endpoints is None:
            endpoints = []

        if _has_valid_schema(host):
            if host.endswith("/"):
                host = host[:-1]
            self._host = host
        else:
            raise ValueError("Host attribute must start either with 'ws://' or 'wss://'")

        self._callback = callback
        self._port = port
        self._running = False
        self._ws_base_url = ws_base_url
        self._endpoints = endpoints
        self._mode = WSMode.DATA_MODE
        self._node_id = self._get_node_id()
        self._ws: WebSocketApp = WebSocketApp(
            f"{self._host}:{self._port}{self._ws_base_url}",
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
        )

    def start_client(self) -> None:
        """
        This will basically only call WebSocketApp.run_forever(). Since this blocks
        until the connection ends, dispatch this call to a separate thread in case
        there is code that should run afterwards.
        """
        self._ws.run_forever()

    def stop_client(self) -> None:
        if self._ws.keep_running:
            pass
        self._ws.close()

    @property
    def node_id(self) -> str:
        return self._node_id

    @property
    def mode(self) -> WSMode:
        return self._mode

    @mode.setter
    def mode(self, val: WSMode) -> None:
        if val == WSMode.DATA_MODE and self._mode != WSMode.DATA_MODE:
            self._ws.send(data=self._DATA_MODE)
        elif val == WSMode.JSON_MODE and self._mode != WSMode.JSON_MODE:
            self._ws.send(data=self._JSON_MODE)
        self._mode = val

    def send_data(
        self,
        destination: str,
        data: bytes,
        source: Optional[str] = None,
        delivery_notification: bool = False,
        lifetime: int = 24 * 3600 * 1000,
    ):
        if source is None:
            source = self._node_id

        bundle_dict: dict = {
            "src": source,
            "dst": destination,
            "delivery_notification": delivery_notification,
            "lifetime": lifetime,
            "data": data,
        }

        payload: bytes
        if self._mode == WSMode.DATA_MODE:
            payload = cbor.dumps(bundle_dict)
        else:
            try:
                # encode in base64 and translate to str equivalent according to spec:
                # https://github.com/dtn7/dtn7-rs/blob/9b166/doc/http-client-api.md
                bundle_dict["data"] = b64encode(bundle_dict["data"]).decode("utf-8")
            except TypeError as e:
                raise TypeError(f"Argument data must be of type 'bytes': {e}")
            json_str = json.dumps(bundle_dict)
            payload = json_str.encode()
        self._ws.send(payload, opcode=ABNF.OPCODE_BINARY)

    def subscribe(self, endpoint: str):
        self._ws.send(data=f"{self._SUBSCRIBE_ENDPOINT} {endpoint}")

    def _on_open(self, ws: WebSocketApp) -> None:
        self._ws.send(data=self._DATA_MODE)
        for eid in self._endpoints:
            ws.send(data=f"{self._SUBSCRIBE_ENDPOINT} {eid}")
        self._running = True

    def _on_message(self, ws: WebSocketApp, msg: Any) -> None:
        # print(f"{msg}")
        # _log(msg)
        # self.messages.append(msg)
        self._callback(msg)

    def _on_error(self, ws: WebSocketApp, error) -> None:
        print(f"{error}")

    def _on_close(self, ws: WebSocketApp, status_code, msg) -> None:
        # print(f"{status_code}, {msg}")
        self._running = False
        print("Connection closed")

    def _get_node_id(self) -> str:
        short_ws: WebSocket = WebSocket()
        short_ws.connect(
            url=f"{self._host}:{self._port}{self._ws_base_url}",
        )
        short_ws.send(self._NODE_ID_ENDPOINT)
        resp: str = short_ws.recv()
        if not resp.startswith("200 node:"):
            raise RuntimeError("Node ID could not be determined.")
        short_ws.close()
        return resp.split(":", maxsplit=1)[1].strip()

    @property
    def running(self) -> bool:
        return self._running



import sys

from datetime import datetime, timedelta, timezone

REF_DT = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0, tzinfo=timezone.utc)
"""
A DTN time is an unsigned integer indicating the number of milliseconds that have elapsed since the
DTN Epoch, 2000-01-01 00:00:00 +0000 (UTC). DTN time is not affected by leap seconds.
(RFC 9171, Section 4.2.6.)
"""

RUNNING_MICROPYTHON = sys.implementation.name == "micropython"


def from_dtn_timestamp(timestamp: int) -> datetime:
    """
    Converts a DTN timestamp to a Python datetime object
    :param timestamp: DTN timestamp
    :return: a Python datetime object representing the DTN timestamp
    """
    return REF_DT + timedelta(milliseconds=timestamp)


def to_dtn_timestamp(dt: datetime = None) -> int:
    if dt is None:
        dt = datetime.now(timezone.utc)

    """
    Converts a Python datetime object into a DTN timestamp
    :param dt:
    :return:
    """
    return int((dt - REF_DT).total_seconds() * 1000)  # cutoff beyond milliseconds, no rounding here



import json
from typing import ClassVar, List, Optional, Union

from py_dtn7 import Bundle
from py_dtn7.utils import RUNNING_MICROPYTHON

if not RUNNING_MICROPYTHON:
    import requests
else:
    import urequests as requests


def has_valid_schema(host: str):
    return host.startswith("https://") or host.startswith("http://")


class DTNRESTClient:
    # API endpoints
    DOWNLOAD_ENDPOINT: ClassVar[str] = "/download"
    ENDPOINT_ENDPOINT: ClassVar[str] = "/endpoint"
    REGISTER_ENDPOINT: ClassVar[str] = "/register"
    SEND_ENDPOINT: ClassVar[str] = "/send"
    PUSH_ENDPOINT: ClassVar[str] = "/push"
    UNREGISTER_ENDPOINT: ClassVar[str] = "/unregister"
    STATUS_BUNDLES: ClassVar[str] = "/status/bundles"
    STATUS_FILTER_BUNDLES: ClassVar[str] = "/status/bundles/filtered"
    STATUS_EIDS: ClassVar[str] = "/status/eids"
    STATUS_INFO: ClassVar[str] = "/status/info"
    STATUS_NODEID: ClassVar[str] = "/status/nodeid"
    STATUS_PEERS: ClassVar[str] = "/status/peers"
    STATUS_STORE: ClassVar[str] = "/status/store"

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        if port is None:
            port = 3000
        if host is None:
            host = "http://localhost"

        if has_valid_schema(host):
            if host.endswith("/"):
                host = host[:-1]
            self._host = host
        else:
            raise ValueError("Host attribute must start either with 'http://' or 'https://'")

        self._port = port
        self._nodeid = self._get_nodeid()

    def send(
        self,
        payload: Union[bytes, dict, str],
        destination: Optional[str] = None,
        peer_name: Optional[str] = None,
        endpoint: Optional[str] = None,
        lifetime: Optional[int] = None,
        encoding: Optional[str] = None,
    ) -> requests.Response:
        dst: str = ""

        if destination is not None:
            dst = destination
        elif peer_name is not None and endpoint is not None:
            peers = self.peers
            if peer_name in peers.keys():
                dst = f"dtn:{peers[peer_name]['eid'][1]}{endpoint}"
            else:
                known_peers: str = ", ".join([k for k in peers.keys()])
                raise ValueError(f"Unknown peer name passed. Known peers: {known_peers}")
        url: str = f"{self._host}:{self._port}{self.SEND_ENDPOINT}?dst={dst}"
        if lifetime is not None:
            url += f"&lifetime={lifetime}"
        if type(payload) is str:
            if encoding is None:
                encoding = "utf-8"
            response: requests.Response = requests.post(
                url=url, data=payload.encode(encoding=encoding)
            )
        elif type(payload) is dict:
            response: requests.Response = requests.post(url=url, json=payload)
        elif type(payload) is bytes:
            response: requests.Response = requests.post(url=url, data=payload)
        else:
            raise ValueError("Payload must by of type 'bytes', 'dict' or 'str'.")

        return response

    def push(self, bundle: bytes):
        url = "{}:{}{}".format(self._host, self._port, self.PUSH_ENDPOINT)
        return requests.post(url=url, data=bundle)

    def register(self, endpoint: str) -> requests.Response:
        response: requests.Response = requests.get(
            url=f"{self._host}:{self._port}{self.REGISTER_ENDPOINT}?{endpoint}"
        )
        result: str = response.content.decode("utf-8").lower()
        if "registered" not in result or endpoint.lower() not in result:
            raise RuntimeError(f'Something went wrong, endpoint "{endpoint}" not registered')
        return response

    def unregister(self, endpoint: str) -> requests.Response:
        response: requests.Response = requests.get(
            url=f"{self._host}:{self._port}{self.UNREGISTER_ENDPOINT}?{endpoint}"
        )
        result: str = response.content.decode("utf-8").lower()
        if "unregistered" not in result or endpoint not in result:
            raise RuntimeError(f'Something went wrong, endpoint "{endpoint}" not unregistered')
        return response

    def get_all_bundles(self) -> List[Bundle]:
        """
        Gets all bundles as Bundle objects from the DTNd store. This can cause very much traffic
        so be sure to use with care.
        :return: all Bundles that are in the DTNd store
        """
        return [
            Bundle.from_cbor(bundle.content)
            for bundle in [
                requests.get(url=f"{self._host}:{self._port}{self.DOWNLOAD_ENDPOINT}?{burl}")
                for burl in self._raw_bundles
            ]
        ]

    def get_filtered_bundles(self, address_part_criteria: str) -> List[str]:
        """

        :param address_part_criteria:
        :return:
        """
        try:
            bundles: List[str] = json.loads(
                requests.get(
                    url="{}:{}{}?addr={}".format(
                        self._host, self._port, self.STATUS_FILTER_BUNDLES, address_part_criteria
                    )
                ).content
            )
        except json.decoder.JSONDecodeError:
            # either no or invalid response received so just return empty list
            return []

        # return [
        #     Bundle.from_cbor(bundle.content)
        #     for bundle in [
        #         rq.get(url=f"{self._host}:{self._port}{self.DOWNLOAD_ENDPOINT}?{burl}")
        #         for burl in bundles
        #     ]
        # ]
        return bundles

    def fetch_endpoint(self, endpoint: str = None) -> bytes:
        if endpoint is None:
            endpoint = self._nodeid
        return requests.get(
            url=f"{self._host}:{self._port}{self.ENDPOINT_ENDPOINT}?{endpoint}"
        ).content

    def download(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        bundle_id: Optional[str] = None,
        node_id: Optional[str] = None,
        peer_name: Optional[str] = None,
        time: Optional[int] = None,
        seq: Optional[int] = None,
    ) -> bytes:
        if bundle_id is None:
            if time is None or seq is None:
                raise ValueError(
                    "In absence of a bundle_id, time and sequence number must be provided."
                )
            if node_id is None:
                if peer_name is None:
                    node_id = self._nodeid
                else:
                    peers = self.peers
                    if peer_name in peers.keys():
                        node_id = f"dtn:{peers[peer_name]['eid'][1]}"
                    else:
                        raise ValueError("Peer not found")
            else:
                if node_id[-1] != "/":
                    node_id = f"{node_id}/"
            bundle_id = f"{node_id}-{time}-{seq}"

        if host is None:
            if port is None:
                host = self._host
                port = self._port
            else:
                raise ValueError("Host and port must either both be defined or undefined.")
        else:
            if not has_valid_schema(host):
                raise ValueError("Host attribute must start either with 'http://' or 'https://'")
            if host[-1] == "/":
                host = host[:-1]

        return requests.get(url=f"{host}:{port}{self.DOWNLOAD_ENDPOINT}?{bundle_id}").content

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def endpoints(self) -> list:
        eps: list = json.loads(
            requests.get(url=f"{self._host}:{self._port}{self.STATUS_EIDS}").content
        )
        # eps = list(map(lambda ep: ep.rsplit("/", 1)[1], eps))
        return eps

    @property
    def bundles(self) -> List[str]:
        return self._raw_bundles

    @property
    def _raw_bundles(self) -> List[str]:
        return json.loads(
            requests.get(url=f"{self._host}:{self._port}{self.STATUS_BUNDLES}").content
        )

    @property
    def store(self) -> list:
        return json.loads(requests.get(url=f"{self._host}:{self._port}{self.STATUS_STORE}").content)

    @property
    def info(self) -> dict:
        return json.loads(requests.get(url=f"{self._host}:{self._port}{self.STATUS_INFO}").content)

    @property
    def peers(self) -> dict:
        return json.loads(requests.get(url=f"{self._host}:{self._port}{self.STATUS_PEERS}").content)

    @property
    def node_id(self) -> str:
        return self._nodeid

    def __str__(self):
        return self.__repl__()

    def __repl__(self):
        return f"<DTNClient@{self._host}:{self._port}, node ID: {self._nodeid}>"

    def _get_nodeid(self) -> Optional[str]:
        return requests.get(url=f"{self._host}:{self._port}{self.STATUS_NODEID}").content.decode(
            "utf-8"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['py_dtn7']

package_data = \
{'': ['*']}

install_requires = \
['cbor2>=5.4.3,<6.0.0',
 'requests>=2.27.1,<3.0.0',
 'websocket-client>=1.3.2,<2.0.0']

setup_kwargs = {
    'name': 'py-dtn7',
    'version': '0.3.0a1',
    'description': '',
    'long_description': '# py-dtn7 (Work in Progress -- don\'t use yet)\n\n[![Licence AGPL-3.0](https://img.shields.io/github/license/teschmitt/py-dtn7)](LICENSE)\n\nA Python wrapper library for the DTN7 REST and WebSocket API of [dtn7-rs](https://github.com/dtn7/dtn7-rs).\nThe library includes a fully spec compliant `Bundle` type (but without fragment and CRC support), which allows full\nbundle creation and (de-)serialization.\n\nThe `bundle.py`, `dtn_rest_client.py`, `utils.py` modules/files are Micropython compatible.\nBut, this requires a bit of manual library management.\nPlease refer to the Micropython installation guide below.\n\n### Getting Started\n\nTo use `py-dtn7` in your project, simply install it from PyPI with Poetry:\n\n```shell\n$ poetry install --no-dev py_dtn7\n```\n\n### Development\n\nThis is very much a work-in-progress and by far not complete. The Bundle\nimplementation is very rudimentary and does not support any blocks other\nthan Primary and Payload.\n\nTo generate the API documentation use `pdoc`:\n\n```shell\n$ pdoc ./py_dtn7 --output-directory ./docs\n```\n\n... or check out [py-dtn7.readthedocs.org](https://py-dtn7.readthedocs.org)\n\n\n## Quickstart\n\n```pycon\n>>> from py_dtn7 import DTNRESTClient\n>>> client = DTNRESTClient(host="http://localhost", port=3000)\n>>> d.peers\n{\'box1\': {\'eid\': [1, \'//box1/\'], \'addr\': {\'Ip\': \'10.0.0.42\'}, \'con_type\': \'Dynamic\', \'period\': None, \'cla_list\': [[\'MtcpConvergenceLayer\', 16162]], \'services\': {}, \'last_contact\': 1653316457}}\n>>> d.info\n{\'incoming\': 0, \'dups\': 0, \'outgoing\': 0, \'delivered\': 3, \'broken\': 0}\n```\n\nWhen sending a bundle to a known peer, we can simply supply the peer name and endpoint,\notherwise we use the complete URI:\n\n```pycon\n>>> d.send(payload={"body": "This will be transferred as json"}, peer_name="box1", endpoint="info")\n<Response [200]>\n>>> r = d.send(payload="Is there anybody out there?", destination="dtn://greatunkown/incoming")\n>>> r.content.decode("utf-8")\n\'Sent payload with 27 bytes\'\n```\n\n## Micropython Installation Guide\n\nTo be extended:\n\nThe dummy libraries `__future__.py`, `abc.py`, `typing.py`, the [micropython-cbor](https://github.com/alexmrqt/micropython-cbor/) library (specifically the `cbor.py` module/file) and `urequests` as well as `datetime` are needed:\n\n```shell\n$ mpremote mip install urequests\n$ mpremote mip install datetime\n```\n',
    'author': 'Thomas Schmitt',
    'author_email': 't.e.schmitt@posteo.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/teschmitt/py-dtn7',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
